################################################################################
### Cache.pm (Yggdrasil::Cache)
### Manages titles and update times for a directory of HTML files
###
### Copyright (C) 2003 Morten Wulff <wulff@ratatosk.net>
###
### This program is free software; you can redistribute it and/or
### modify it under the terms of the GNU General Public License
### as published by the Free Software Foundation; either version 2
### of the License, or (at your option) any later version.
###
### This program is distributed in the hope that it will be useful,
### but WITHOUT ANY WARRANTY; without even the implied warranty of
### MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
### GNU General Public License for more details.
###
### You should have received a copy of the GNU General Public License
### along with this program; if not, write to the Free Software
### Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
################################################################################

### REVISIONS
###	1.0.0 : 2003-04-11 : Morten Wulff <wulff@ratatosk.net>
###	1.1.0 : 2003-04-24 : Morten Wulff <wulff@ratatosk.net>
###			Improved interface, easier to include/exclude files
###	1.1.1 : 2003-04-26 : Morten Wulff <wulff@ratatosk.net>
###			Minor code clean-up
###	1.1.2 : 2003-04-27 : Morten Wulff <wulff@ratatosk.net>
###			Fixed problem in _unlink_cache
###	1.2.0 : 2003-05-22 : Morten Wulff <wulff@ratatosk.net>
###			Fixed title() cache update behavior

### TODO
### _process_file should use grep to check if a file is in the html_indices array

package Yggdrasil::Cache;

use strict;
use warnings;

$|++;

use Carp;
use Fcntl;
use File::Basename;
use File::Find;
use File::Path;
use SDBM_File;

################################################################################
### INIT
################################################################################

our $AUTOLOAD;

my %Fields = (
	include => ['\.html$', '\.shtml$'], ### include files matching these regexp's (include is applied before exclude) 
	exclude => ['^_', '\.tmpl$'],       ### exclude files/paths matching these regexp's

	input_dir => '',   ### where to start crawling
	data_dir  => '',   ### store SDBM files here
	dir_mode  => 0644, ### dir create mode for mkpath

	html_indices => [qw(index.html index.shtml)],  ### should match your apache config (are stripped from generated URL's)

	default_title => 'No Title', ### return for files not in database

	force   => 0, ### force rebuild of cache (old database files are deleted)
	verbose => 0  ### level of verbosity
);

################################################################################
### PUBLIC METHODS
################################################################################

################################################################################
sub new {
	my $proto = shift;
	my $class = ref($proto) || $proto;

	my %parameters = @_;

	my $self  = {
		_permitted => \%Fields
	};

	foreach my $element (keys %Fields) {
		$self->{$element} = $parameters{$element} || $Fields{$element};
	}

	### check for presence of input directory
	unless ( -d $self->{input_dir} ) { croak "Input directory '$self->{input_dir}' doesn't exist"; }

	bless ($self, $class);

	$self->_init(); ### tie cache files and perform updates

	return $self;
}

################################################################################
sub size {
	my $self = shift;

	return scalar(keys %{$self->{update}});
}

################################################################################
sub title {
	my $self     = shift;
	my $filename = shift;

	unless (exists $self->{title}->{$filename}) {
		if (-e $filename) {
			SWITCH: {
				-d $filename && do { ### $filename is a directory
					my $directory = $filename;
					foreach my $html_index (@{$self->{html_indices}}) {
						my $filename = join '/', $directory, $html_index;
						if (-f $filename) {
							$self->_process_file($filename, $directory);
						}
					}
					last SWITCH;
				};
				-f $filename && do { ### $filename is a plain file
					my $directory = dirname($filename);
					$self->_process_file($filename, $directory);
					last SWITCH;
				};

				### fallthrough
				carp "File '$filename' isn't a plain file or a dirctory";
			}
		}
		else {
			carp "File '$filename' doesn't exist";
			return;
		}
	}

	return exists $self->{title}->{$filename}
		? $self->{title}->{$filename}
		: $self->{default_title};
}

################################################################################
sub update {
	my $self     = shift;
	my $filename = shift;

	unless (exists $self->{update}->{$filename}) {
		if (-e $filename) {
			my $directory = dirname($filename);
			$self->_process_file($filename, $directory);
		}
		else {
			carp "File '$filename' doesn't exist";
			return;
		}
	}

	return exists $self->{update}->{$filename} ? $self->{update}->{$filename} : 0;
}

################################################################################
### PRIVATE METHODS
################################################################################

################################################################################
### http://perlmonks.thepen.com/109068.html
sub _bind_args {
	my $self = shift;
	my ($sub, @args) = @_;

	return sub {$sub->(@args, @_);};
}

################################################################################
sub _create_dir {
	my $self = shift;

	$self->{force} = 1; ### force cache build

	print "Creating data directory '$self->{data_dir}'\n";

	mkpath($self->{data_dir}, 0, $self->{dir_mode}) or croak "Cannot create directory '$self->{data_dir}': $!";
}

################################################################################
sub _init {
	my $self = shift;

	### create cache directory if it doesn't exist
	$self->_create_dir() unless (-d $self->{data_dir});

	### force cache build if data files are missing
	unless (
		-e "$self->{data_dir}/title.dir"  &&
		-e "$self->{data_dir}/title.pag"  &&
		-e "$self->{data_dir}/update.dir" &&
		-e "$self->{data_dir}/update.pag"
	) {
		print "Forcing cache creation\n";
		$self->{force} = 1;
	}

	$self->_unlink_cache() if $self->{force};
	$self->_tie_cache();

	if ($self->{force}) {
		print "Creating cache in '$self->{data_dir}'... ";
		print "\n" if ($self->{verbose});

		find(
			{
				wanted      => $self->_bind_args(\&_process, $self),
				preprocess  => $self->_bind_args(\&_pre_process, $self)
			},
			$self->{input_dir}
		);

		print "Done!\n" unless ($self->{verbose});
	};
}

################################################################################
sub _include_in_cache {
	my $self = shift;
	my $file = shift;

	return grep {$file =~ /$_/} @{$self->{include}};
}

################################################################################
sub _pre_process {
	my $self = shift;

	my @files = sort { $a cmp $b } @_;

	### remove unwanted elements from the file list
	@files = grep {
		my $file = $_;
		my $path = join '/', $File::Find::dir, $file;
		!/^\.{1,2}/ &&
		!grep {$file =~ /$_/ || $path =~ /$_/} @{$self->{exclude}}
	} @files;

	return @files;
}

################################################################################
sub _process {
	my $self = shift;

	return if -d;

	$self->_process_file($File::Find::name, $File::Find::dir);
}

################################################################################
sub _process_file {
	my $self = shift;
	my ($filename, $directory) = @_;

	if ($self->_include_in_cache($filename)) {
		open(IN, $filename) or croak "Cannot open $filename: $!\n";
		while ( my $line = <IN> ) {
			if ( my ($title) = ($line =~ /<h1.*?>(.*)<\/h1>/) ) {
				$self->{title}->{$filename}  = $title;
				$self->{title}->{$directory} = $title if ($filename =~ /index/);
				last;
			}
		}
		close(IN);

		print "$filename\n" if $self->{verbose};
	}

	$self->{update}->{$filename} = (stat($filename))[9];
}

################################################################################
sub _tie_cache {
	my $self = shift;

	tie(my %title,  'SDBM_File', "$self->{data_dir}/title",  O_RDWR|O_CREAT, 0666) or croak "Cannot tie SDBM file 'title': $!\n";
	tie(my %update, 'SDBM_File', "$self->{data_dir}/update", O_RDWR|O_CREAT, 0666) or croak "Cannot tie SDBM file 'update': $!\n";

	$self->{title}  = \%title;
	$self->{update} = \%update;
}

################################################################################
sub _unlink_cache {
	my $self = shift;

	-e "$self->{data_dir}/title.dir" && do { unlink "$self->{data_dir}/title.dir" or croak "Cannot unlink 'title.dir': $!\n"; };
	-e "$self->{data_dir}/title.pag" && do { unlink "$self->{data_dir}/title.pag" or croak "Cannot unlink 'title.pag': $!\n"; };

	-e "$self->{data_dir}/update.dir" && do { unlink "$self->{data_dir}/update.dir" or croak "Cannot unlink 'update.dir': $!\n"; };
	-e "$self->{data_dir}/update.pag" && do { unlink "$self->{data_dir}/update.pag" or croak "Cannot unlink 'update.pag': $!\n"; };
}

################################################################################
sub AUTOLOAD {
	my $self = shift;
	my $type = ref($self) or croak "$self is not an object";

	my $name = $AUTOLOAD;
	$name =~ s/.*://;

	unless (exists $self->{_permitted}->{$name} ) {
		croak "Can't access `$name' field in class $type";
	}

	if (@_) {
		return $self->{$name} = shift;
	}
	else {
		return $self->{$name};
	}
}

################################################################################
END {
	my $self = shift;

	untie(%{$self->{title}})  or warn "Cannot untie title cache\n";
	untie(%{$self->{update}}) or warn "Cannot untie update cache\n";
}

1;

################################################################################
### DOCUMENTATION
################################################################################

__END__

=head1 NAME

Yggdrasil::Cache - Manages titles and update times for a directory of HTML files

=head1 SYNOPSIS

=head1 DESCRIPTION

=head1 BUGS

No known bugs.

=head1 AUTHOR

Morten Wulff <wulff@ratatosk.net>

Copyright (C) 2003 Morten Wulff

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

=head1 VERSION

Version 1.1.2
